/*
 *
 * Compiz copy mode plugin
 *
 * Copyright : (C) 2007 by Dennis Kasprzyk
 * E-mail    : onestone@beryl-project.org
 *
 *
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License
 * as published by the Free Software Foundation; either version 2
 * of the License, or (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 */

/*
 * This plugin uses the "copy to texture" taken from Luminocity
 * http://live.gnome.org/Luminocity
 */

#ifdef HAVE_CONFIG_H
#include <config.h>
#endif

#define _GNU_SOURCE
#include <string.h>
#include <stdlib.h>
#include <math.h>
#include <sys/time.h>

#include <X11/Xlib.h>
#include <X11/Xatom.h>
#include <X11/Xproto.h>
#include <X11/extensions/XShm.h>

#include <sys/shm.h>
#include <sys/ipc.h>

#include <compiz.h>

#define FORCE_COPY FALSE
#define USE_SHM FALSE

#define SHM_SIZE 4096*2048*4

static CompMetadata copyMetadata;

static int displayPrivateIndex;

typedef struct _CopyTexture CopyTexture;

typedef struct _CopySubTexture
{
    int width;
    int height;
    int offX;
    int offY;
    GLuint name;

    GLfloat  *vertices;
    int      vertexSize;
    GLushort *indices;
    int      indexSize;
    int      vCount;
    int      indexCount;

} CopySubTexture;

struct _CopyTexture {
    int width;
    int height;
    int depth;
    Damage xdamage;
    Drawable damageXid;
    Bool damaged;
    BOX  damage;

    CopySubTexture *st;
    int numSubTex;
    int activeSubTex;

    CopyTexture *next;
};

typedef struct _CopyDisplay {

    HandleEventProc handleEvent;

    int	screenPrivateIndex;

    XShmSegmentInfo shmInfo;
    Bool noShm;
    int maxTexSize;

    CopyTexture *textures;
} CopyDisplay;


typedef struct _CopyScreen {
    FiniTextureProc              finiTexture;
    BindPixmapToTextureProc      bindPixmapToTexture;
    ReleasePixmapFromTextureProc releasePixmapFromTexture;
    EnableTextureProc            enableTexture;
    DisableTextureProc           disableTexture;
    AddWindowGeometryProc        addWindowGeometry;
    DrawWindowTextureProc drawWindowTexture;
} CopyScreen;


#define GET_COPY_DISPLAY(d)				     \
    ((CopyDisplay *) (d)->privates[displayPrivateIndex].ptr)

#define COPY_DISPLAY(d)			   \
    CopyDisplay *cd = GET_COPY_DISPLAY (d)

#define GET_COPY_SCREEN(s, cd)					 \
    ((CopyScreen *) (s)->privates[(cd)->screenPrivateIndex].ptr)

#define COPY_SCREEN(s)							\
    CopyScreen *cs = GET_COPY_SCREEN (s, GET_COPY_DISPLAY (s->display))

#define COPY_TEXTURE(t) \
    CopyTexture *ct = (t)? (CopyTexture*) (t)->private : NULL

static void
copyAddWindowGeometry (CompWindow  *w,
		   CompMatrix  *matrix,
		   int	       nMatrix,
		   CompTexture *texture,
		   Region      region,
		   Region      clip)
{
    COPY_SCREEN(w->screen);
    COPY_TEXTURE(texture);
    Region reg;
    XRectangle rect;
    CompMatrix *mCopy;
    Bool new = FALSE;
    int i,j;

    if (!ct || !ct->numSubTex || !nMatrix)
    {
	UNWRAP (cs, w->screen, addWindowGeometry);
	(*w->screen->addWindowGeometry) (w, matrix, nMatrix, texture,
					region, clip);
	WRAP (cs, w->screen, addWindowGeometry, copyAddWindowGeometry);
	return;
    }

    if (!w->vCount)
	new = TRUE;

    if (w->vertexSize)
    {
	free (w->vertices);
	w->vertices = NULL;
	w->vertexSize = 0;
	w->vCount = 0;
    }
    if (w->indexSize)
    {
	free (w->indices);
	w->indices = NULL;
	w->indexSize = 0;
	w->indexCount = 0;
    }

    reg = XCreateRegion();

    mCopy = malloc(sizeof (CompMatrix) * nMatrix);

    for (i = 0; i < ct->numSubTex; i++)
    {
	CopySubTexture *st = &ct->st[i];
	XSubtractRegion (&emptyRegion, &emptyRegion, reg);
	memcpy (mCopy, matrix, sizeof (CompMatrix) * nMatrix);

	// we use always TEXTURE_RECTANGLE
	if (mCopy[0].xx < 1 && mCopy[0].xy < 1)
	{
	    rect.x = 0;
	    rect.width = w->screen->width;
	}
	else if (mCopy[0].xx == 1 && mCopy[0].xy < 1)
	{
	    rect.x = st->offX - mCopy[0].x0;
	    rect.width = st->width;
	}
	else if (mCopy[0].xx < 1 && mCopy[0].xy == 1)
	{
	    rect.x = st->offY + mCopy[0].x0;
	    rect.width = st->height;
	}

	if (mCopy[0].yy < 1 && mCopy[0].yx < 1)
	{
	    rect.y = 0;
	    rect.height = w->screen->height;
	}
	else if (mCopy[0].yy == 1 && mCopy[0].yx < 1)
	{
	    rect.y = st->offY - mCopy[0].y0;
	    rect.height = st->height;
	}
	else if (mCopy[0].yy < 1 && mCopy[0].yx == 1)
	{
	    rect.y = st->offX - mCopy[0].y0;
	    rect.height = st->width;
	}

	XUnionRectWithRegion(&rect, reg, reg);

	XIntersectRegion (reg, region, reg);

        for (j = 0; j < nMatrix; j++)
	{
	   mCopy[j].x0 -= (st->offX * mCopy[j].xx);
           mCopy[j].y0 -= (st->offY * mCopy[j].yy);
	}

	w->vertices = st->vertices;
	w->vertexSize = st->vertexSize;
	w->vCount = (new) ? 0 : st->vCount;
	w->indices = st->indices;
	w->indexSize = st->indexSize;
	w->indexCount = (new) ? 0 : st->indexCount;

	UNWRAP (cs, w->screen, addWindowGeometry);
	(*w->screen->addWindowGeometry) (w, mCopy, nMatrix, texture,
					reg, clip);
	WRAP (cs, w->screen, addWindowGeometry, copyAddWindowGeometry);

	st->vertices = w->vertices;
	st->vertexSize = w->vertexSize;
	st->vCount = w->vCount;
	st->indices = w->indices;
	st->indexSize = w->indexSize;
	st->indexCount = w->indexCount;
    }

    w->vertices = 0;
    w->vertexSize = 0;
    w->vCount = 1;
    w->indices = NULL;
    w->indexSize = 0;
    w->indexCount = 0;


    free (mCopy);
    XDestroyRegion(reg);
}

static void
copyDrawWindowTexture (CompWindow		*w,
		   CompTexture		*texture,
		   const FragmentAttrib	*attrib,
		   unsigned int		mask)
{
    COPY_SCREEN(w->screen);
    COPY_TEXTURE(texture);
    int i;

    if (!ct || !ct->numSubTex)
    {
	UNWRAP (cs, w->screen, drawWindowTexture);
	(*w->screen->drawWindowTexture) (w, texture, attrib, mask);
	WRAP (cs, w->screen, drawWindowTexture, copyDrawWindowTexture);
	return;
    }

    for (i = 0; i < ct->numSubTex; i++)
    {
	CopySubTexture *st = &ct->st[i];
	ct->activeSubTex = i;
	w->vertices = st->vertices;
	w->vertexSize = st->vertexSize;
	w->vCount = st->vCount;
	w->indices = st->indices;
	w->indexSize = st->indexSize;
	w->indexCount = st->indexCount;

	if (!w->vCount)
	    continue;

        UNWRAP (cs, w->screen, drawWindowTexture);
	(*w->screen->drawWindowTexture) (w, texture, attrib, mask);
	WRAP (cs, w->screen, drawWindowTexture, copyDrawWindowTexture);
    }

    w->vertices = NULL;
    w->vertexSize = 0;
    w->vCount = 0;
    w->indices = NULL;
    w->indexSize = 0;
    w->indexCount = 0;

}

Bool
copyBindPixmapToTexture (CompScreen  *s,
		     CompTexture *t,
		     Pixmap	 pixmap,
		     int	 width,
		     int	 height,
		     int	 depth)
{
    COPY_DISPLAY(s->display);
    COPY_SCREEN(s);
    COPY_TEXTURE(t);

    if (width <= cd->maxTexSize || height <= cd->maxTexSize || FORCE_COPY)
    {
	Bool status;
	UNWRAP (cs, s, bindPixmapToTexture);
        status = (*s->bindPixmapToTexture) (s, t, pixmap, width, height, depth);
        WRAP (cs, s, bindPixmapToTexture, copyBindPixmapToTexture);
	if (status)
	    return status;
    }

    if (!ct)
    {
	ct = calloc (1, sizeof(CopyTexture));
	t->private = (void*)ct;
	CopyTexture *ot = cd->textures;
	cd->textures = ct;
	ct->next = ot;
	ct->damageXid = None;
	ct->st = NULL;
        ct->numSubTex = 0;
	ct->activeSubTex = 0;
    }

    if (!t->name || ct->width != width || ct->height != height)
    {
	if (ct->xdamage)
	{
	    XDamageDestroy (s->display->display, ct->xdamage);
	}
	if (!t->name)
	{
	   glGenTextures(1, &t->name);
	}
	Drawable dXid = pixmap;
	CompWindow *w = s->windows;
	while (w)
	{
	    if (w->pixmap == pixmap)
	    {
		dXid = w->id;
	    }
	    w = w->next;
	}
	ct->damageXid = dXid;
	ct->xdamage = XDamageCreate(s->display->display, dXid,
				   XDamageReportBoundingBox);

	t->pixmap = pixmap;
	ct->width = width;
	ct->height = height;
	ct->depth = depth;
	ct->damage.x1 = 0;
	ct->damage.y1 = 0;
	ct->damage.x2 = width;
	ct->damage.y2 = height;
	ct->damaged = TRUE;

	t->target = GL_TEXTURE_RECTANGLE_ARB;
        t->filter = GL_NEAREST;
	t->wrap = GL_CLAMP_TO_EDGE;
	t->matrix.xx = 1.0f;
	t->matrix.yy = 1.0f;
	t->matrix.y0 = 0;


	if (width > cd->maxTexSize || height > cd->maxTexSize)
	{
	    int nX = ceil((float) width / (float)cd->maxTexSize);
	    int nY = ceil((float) height / (float)cd->maxTexSize);
	    int i,j,k;
	    int tw,th;

	    if (ct->numSubTex)
	    {
		for (k = 0; k < ct->numSubTex; k++)
		{
		    glDeleteTextures (1, &ct->st[k].name);
		    if (ct->st[k].vertexSize)
			free (ct->st[k].vertices);
		    if (ct->st[k].indexSize)
			free (ct->st[k].indices);
		}
		free (ct->st);
		ct->numSubTex = 0;
	    }

	    ct->numSubTex = nX * nY;
	    ct->st = calloc (1, sizeof(CopySubTexture) * ct->numSubTex);
	    k = 0;
	    tw = floor((float) width / (float)nX);
	    th = floor((float) height / (float)nY);
	    for (j = 0; j < nY; j++)
		for (i = 0; i < nX; i++)
		{
		    ct->st[k].width = (i == nX-1)? width - (i * tw): tw;
		    ct->st[k].height = (j == nY-1)? height - (j * th): th;
		    ct->st[k].offX = i * tw;
		    ct->st[k].offY = j * th;

                    glGenTextures(1, &ct->st[k].name);

		    glBindTexture(t->target, ct->st[k].name);

		    glTexParameteri(t->target, GL_TEXTURE_MIN_FILTER, GL_NEAREST);
		    glTexParameteri(t->target, GL_TEXTURE_MAG_FILTER, GL_NEAREST);

		    glTexParameteri(t->target, GL_TEXTURE_WRAP_S, GL_CLAMP_TO_EDGE);
		    glTexParameteri(t->target, GL_TEXTURE_WRAP_T, GL_CLAMP_TO_EDGE);

		    if (depth == 32)
			glTexImage2D(GL_TEXTURE_RECTANGLE_ARB, 0, GL_RGBA,
				     ct->st[k].width, ct->st[k].height, 0,
				     GL_BGRA, GL_UNSIGNED_BYTE, 0);
		    else
			glTexImage2D(GL_TEXTURE_RECTANGLE_ARB, 0, GL_RGB,
				     ct->st[k].width, ct->st[k].height, 0,
				     GL_BGRA, GL_UNSIGNED_BYTE, 0);

		    glBindTexture(t->target, 0);
		    k++;
		}
	}
	else
	{
	    glBindTexture(t->target, t->name);

	    glTexParameteri(t->target, GL_TEXTURE_MIN_FILTER, GL_NEAREST);
	    glTexParameteri(t->target, GL_TEXTURE_MAG_FILTER, GL_NEAREST);

	    glTexParameteri(t->target, GL_TEXTURE_WRAP_S, GL_CLAMP_TO_EDGE);
	    glTexParameteri(t->target, GL_TEXTURE_WRAP_T, GL_CLAMP_TO_EDGE);

	    if (depth == 32)
		glTexImage2D(GL_TEXTURE_RECTANGLE_ARB, 0, GL_RGBA,
			     width, height, 0, GL_BGRA, GL_UNSIGNED_BYTE, 0);
	    else
		glTexImage2D(GL_TEXTURE_RECTANGLE_ARB, 0, GL_RGB,
			     width, height, 0, GL_BGRA, GL_UNSIGNED_BYTE, 0);

	    glBindTexture(t->target, 0);
	}

    }
    return TRUE;
}

void copyReleasePixmapFromTexture(CompScreen * s, CompTexture * t)
{
    COPY_SCREEN(s);
    COPY_TEXTURE(t);

    if (ct)
    {
        if (t->name)
	   glDeleteTextures(1, &t->name);
	if (ct->numSubTex)
	{
	    int k;
	    for (k = 0; k < ct->numSubTex; k++)
	    {
		glDeleteTextures (1, &ct->st[k].name);
		if (ct->st[k].vertexSize)
		    free (ct->st[k].vertices);
		if (ct->st[k].indexSize)
		    free (ct->st[k].indices);
	    }
	    free (ct->st);
	    ct->numSubTex = 0;
	}
	t->name = 0;
	t->pixmap = None;
    }
    else
    {
	UNWRAP (cs, s, releasePixmapFromTexture);
	(*s->releasePixmapFromTexture) (s, t);
	WRAP (cs, s, releasePixmapFromTexture, copyReleasePixmapFromTexture);
    }
}


static void
copyToTexture (CompDisplay *d,
	       Pixmap pix,
	       Bool shmMode,
	       int depth,
	       int x,
	       int y,
	       int offX,
	       int offY,
	       int width,
	       int height)
{
    COPY_DISPLAY(d);

    char *addr = 0;
    Pixmap tmpPix;
    Display *dpy = d->display;

    XGCValues gcv;
    GC gc;

    gcv.graphics_exposures = FALSE;
    gcv.subwindow_mode = IncludeInferiors;
    gc = XCreateGC(dpy, pix, GCGraphicsExposures | GCSubwindowMode, &gcv);

    XImage *image = 0;

    if (shmMode)
	tmpPix = XShmCreatePixmap(dpy, pix, cd->shmInfo.shmaddr,
				    &cd->shmInfo, width, height, depth);
    else
	tmpPix = XCreatePixmap(dpy, pix, width, height, depth);

    if (tmpPix == None)
    {
	fprintf(stderr,_("%s: Cannot create Pixmap for copy\n"),
		programName);
	return;
    }

    XCopyArea(dpy, pix, tmpPix, gc, x, y, width, height, 0, 0);
    XSync(dpy, FALSE);

    if (shmMode)
	addr = cd->shmInfo.shmaddr;
    else
    {
	image = XGetImage(dpy, tmpPix, 0, 0, width, height,
			    AllPlanes, ZPixmap);
	if (image)
	    addr = image->data;
    }

    glTexSubImage2D(GL_TEXTURE_RECTANGLE_ARB, 0, offX, offY, width,
		    height, GL_BGRA,
#if IMAGE_BYTE_ORDER == MSBFirst
		    GL_UNSIGNED_INT_8_8_8_8_REV,
#else
                    GL_UNSIGNED_BYTE,
#endif
		    addr);


    XFreePixmap(dpy, tmpPix);
    XFreeGC(dpy, gc);
    if (image)
	XDestroyImage(image);
}

void
copyEnableTexture (CompScreen	 *s,
	       CompTexture	 *t,
	       CompTextureFilter filter)
{
    COPY_SCREEN(s);
    COPY_TEXTURE(t);

    if (!ct)
    {
	UNWRAP (cs, s, enableTexture);
	(*s->enableTexture) (s, t, filter);
	WRAP (cs, s, enableTexture, copyEnableTexture);
	return;
    }

    makeScreenCurrent (s);
    glEnable (t->target);

    if (ct->damaged)
    {
	CompDisplay *d = s->display;
	COPY_DISPLAY(d);

	Bool shmMode = !cd->noShm &&
			(ct->width * ct->height * 4 < SHM_SIZE);

	int width = ct->damage.x2 - ct->damage.x1;
	int height = ct->damage.y2 - ct->damage.y1;
	int x = ct->damage.x1;
	int y = ct->damage.y1;

	if (width <= 0 || height <= 0)
	    return;

	if (ct->numSubTex)
	{
	    int k;
	    for (k = 0; k < ct->numSubTex; k++)
	    {
		int tx = x;
		int ty = y;
		int tw = width;
		int th = height;

		if (x < ct->st[k].offX)
		{
		    tw -= ct->st[k].offX - x;
		    tx = ct->st[k].offX;
		}
		if (tx + tw > ct->st[k].offX + ct->st[k].width)
		    tw = ct->st[k].offX + ct->st[k].width - tx;

		if (y < ct->st[k].offY)
		{
		    th -= ct->st[k].offY - y;
		    ty = ct->st[k].offY;
		}
		if (ty + th > ct->st[k].offY + ct->st[k].height)
		    th = ct->st[k].offY + ct->st[k].height - ty;
		if (tw <= 0 || th <= 0)
		    continue;

		glBindTexture (t->target, ct->st[k].name);
	        copyToTexture (d, t->pixmap, shmMode, ct->depth, tx, ty,
	                   tx - ct->st[k].offX, ty - ct->st[k].offY, tw, th);
	    }
	}
	else
	{
	    glBindTexture (t->target, t->name);
	    copyToTexture (d, t->pixmap, shmMode, ct->depth, x, y,
	                   x, y, width, height);
	}

	ct->damaged = FALSE;
    }

    if (ct->numSubTex)
	glBindTexture (t->target, ct->st[ct->activeSubTex].name);
    else
        glBindTexture (t->target, t->name);



    if (filter == COMP_TEXTURE_FILTER_FAST)
    {
	if (t->filter != GL_NEAREST)
	{
	    glTexParameteri (t->target,
			     GL_TEXTURE_MIN_FILTER,
			     GL_NEAREST);
	    glTexParameteri (t->target,
			     GL_TEXTURE_MAG_FILTER,
			     GL_NEAREST);

	    t->filter = GL_NEAREST;
	}
    }
    else if (t->filter != s->display->textureFilter)
    {
	if (s->display->textureFilter == GL_LINEAR_MIPMAP_LINEAR)
	{
	    if (s->textureNonPowerOfTwo && s->fbo && t->mipmap)
	    {
		glTexParameteri (t->target,
				 GL_TEXTURE_MIN_FILTER,
				 GL_LINEAR_MIPMAP_LINEAR);

		if (t->filter != GL_LINEAR)
		    glTexParameteri (t->target,
				     GL_TEXTURE_MAG_FILTER,
				     GL_LINEAR);

		t->filter = GL_LINEAR_MIPMAP_LINEAR;
	    }
	    else if (t->filter != GL_LINEAR)
	    {
		glTexParameteri (t->target,
				 GL_TEXTURE_MIN_FILTER,
				 GL_LINEAR);
		glTexParameteri (t->target,
				 GL_TEXTURE_MAG_FILTER,
				 GL_LINEAR);

		t->filter = GL_LINEAR;
	    }
	}
	else
	{
	    glTexParameteri (t->target,
			     GL_TEXTURE_MIN_FILTER,
			     s->display->textureFilter);
	    glTexParameteri (t->target,
			     GL_TEXTURE_MAG_FILTER,
			     s->display->textureFilter);

	    t->filter = s->display->textureFilter;
	}
    }

    if (t->filter == GL_LINEAR_MIPMAP_LINEAR)
    {
	if (t->oldMipmaps)
	{
	    (*s->generateMipmap) (t->target);
	    t->oldMipmaps = FALSE;
	}
    }
}


void
copyDisableTexture (CompScreen  *s,
		CompTexture *t)
{
    COPY_SCREEN(s);
    COPY_TEXTURE(t);

    if (!ct)
    {
	UNWRAP (cs, s, disableTexture);
	(*s->disableTexture) (s, t);
	WRAP (cs, s, disableTexture, copyDisableTexture);
	return;
    }
    makeScreenCurrent (s);
    glBindTexture (t->target, 0);
    glDisable (t->target);
}

void
copyFiniTexture (CompScreen  *s,
	         CompTexture *t)
{
    COPY_SCREEN(s);
    COPY_DISPLAY(s->display);
    COPY_TEXTURE(t);
    if (ct && cd->textures)
    {
	if (cd->textures == ct)
	{
	    cd->textures = ct->next;
	}
	else
	{
	    CopyTexture *ot = cd->textures;
	    while (ot)
	    {
		if (ot->next == ct)
		{
		    ot->next = ct->next;
		    break;
		}
		ot = ot->next;
	    }
	}
	XDamageDestroy (s->display->display, ct->xdamage);
        free (ct);
    }
    UNWRAP (cs, s, finiTexture);
    (*s->finiTexture) (s, t);
    WRAP (cs, s, finiTexture, copyFiniTexture);
}

static void
copyHandleEvent (CompDisplay *d,
                XEvent      *event)
{
    COPY_DISPLAY(d);

    if (event->type == d->damageEvent + XDamageNotify)
    {
        XDamageNotifyEvent *de = (XDamageNotifyEvent *) event;

        CopyTexture *t = cd->textures;

	while (t)
	{
	    if (de->drawable == t->damageXid)
        	break;
            t = t->next;
	}

	if (t)
	{
	    if (!t->damaged)
	    {
		t->damage.x1 = de->area.x;
		t->damage.y1 = de->area.y;
		t->damage.x2 = de->area.x + de->area.width;
		t->damage.y2 = de->area.y + de->area.height;
		t->damaged = TRUE;
	    }
	    else
	    {
		t->damage.x1 = MIN(de->area.x, t->damage.x1);
		t->damage.y1 = MIN(de->area.y, t->damage.y1);
		t->damage.x2 = MAX(de->area.x +	de->area.width, t->damage.x2);
		t->damage.y2 = MAX(de->area.y +	de->area.height, t->damage.y2);
	    }
	}
    }

    UNWRAP (cd, d, handleEvent);
    (*d->handleEvent) (d, event);
    WRAP (cd, d, handleEvent, copyHandleEvent);
}


static Bool
copyInitDisplay (CompPlugin  *p,
		 CompDisplay *d)
{
    CopyDisplay *cd;
    CompScreen *s;

    cd = malloc (sizeof (CopyDisplay));
    if (!cd)
	return FALSE;

    cd->screenPrivateIndex = allocateScreenPrivateIndex (d);
    if (cd->screenPrivateIndex < 0)
    {
	free (cd);
	return FALSE;
    }

    Bool noShm = !USE_SHM;
    cd->maxTexSize = MAXSHORT;
    for (s = d->screens; s; s = s->next)
	cd->maxTexSize = MIN(cd->maxTexSize, s->maxTextureSize);

    cd->textures = NULL;

    d->privates[displayPrivateIndex].ptr = cd;

    if (!noShm)
    {
	cd->shmInfo.shmid = shmget(IPC_PRIVATE, SHM_SIZE, IPC_CREAT | 0777);
	if (cd->shmInfo.shmid < 0)
	{
		fprintf(stderr, _("%s: Error in SHM creation\n"), programName);
		noShm = TRUE;
	}

	cd->shmInfo.shmaddr = (char *)shmat(cd->shmInfo.shmid, 0, 0);
	if (cd->shmInfo.shmaddr == ((char *)-1))
	{
		shmctl(cd->shmInfo.shmid, IPC_RMID, 0);
		fprintf(stderr, _("%s: Error in SHM creation\n"), programName);
		noShm = TRUE;
	}
	cd->shmInfo.readOnly = False;

	XShmAttach(d->display, &cd->shmInfo);
    }

    WRAP (cd, d, handleEvent, copyHandleEvent);

    return TRUE;
}

static void
copyFiniDisplay (CompPlugin  *p,
		 CompDisplay *d)
{
    COPY_DISPLAY (d);

    freeScreenPrivateIndex (d, cd->screenPrivateIndex);

    UNWRAP (cd, d, handleEvent);

    free (cd);
}

static Bool
copyInitScreen (CompPlugin *p,
		CompScreen *s)
{
    CopyScreen *cs;

    COPY_DISPLAY (s->display);

    cs = malloc (sizeof (CopyScreen));
    if (!cs)
	return FALSE;

    s->privates[cd->screenPrivateIndex].ptr = cs;

    WRAP (cs, s, bindPixmapToTexture, copyBindPixmapToTexture);
    WRAP (cs, s, releasePixmapFromTexture, copyReleasePixmapFromTexture);
    WRAP (cs, s, disableTexture, copyDisableTexture);
    WRAP (cs, s, enableTexture, copyEnableTexture);
    WRAP (cs, s, finiTexture, copyFiniTexture);
    WRAP (cs, s, addWindowGeometry, copyAddWindowGeometry);
    WRAP (cs, s, drawWindowTexture, copyDrawWindowTexture);


    return TRUE;
}

static void
copyFiniScreen (CompPlugin *p,
		CompScreen *s)
{
    COPY_SCREEN (s);

    UNWRAP (cs, s, bindPixmapToTexture);
    UNWRAP (cs, s, releasePixmapFromTexture);
    UNWRAP (cs, s, disableTexture);
    UNWRAP (cs, s, enableTexture);
    UNWRAP (cs, s, finiTexture);
    UNWRAP (cs, s, addWindowGeometry);
    UNWRAP (cs, s, drawWindowTexture);

    free (cs);
}

static Bool
copyInit (CompPlugin *p)
{
    displayPrivateIndex = allocateDisplayPrivateIndex ();
    if (displayPrivateIndex < 0)
    {
	compFiniMetadata (&copyMetadata);
	return FALSE;
    }

    compAddMetadataFromFile (&copyMetadata, p->vTable->name);
    return TRUE;
}

static void
copyFini (CompPlugin *p)
{
    freeDisplayPrivateIndex (displayPrivateIndex);
    compFiniMetadata (&copyMetadata);
}

static int
copyGetVersion (CompPlugin *plugin,
		int	   version)
{
    return ABIVERSION;
}

static CompMetadata *
copyGetMetadata (CompPlugin *plugin)
{
    return &copyMetadata;
}

CompPluginVTable copyVTable = {
    "copymode",
    copyGetVersion,
    copyGetMetadata,
    copyInit,
    copyFini,
    copyInitDisplay,
    copyFiniDisplay,
    copyInitScreen,
    copyFiniScreen,
    0, /* InitWindow */
    0, /* FiniWindow */
    0,
    0,
    0,
    0,
    0,
    0,
    0,
    0
};

CompPluginVTable *
getCompPluginInfo (void)
{
    return &copyVTable;
}
